/* -*- mode: Java; c-basic-offset: 4 -*- */
/*------------------------------------------------------------------------
 |  FILE            WebPage.java
 |  PART OF         WebWatcher - A web-page update monitor
 |  WRITTEN BY      Sverre H. Huseby <shh@thathost.com>
 +----------------------------------------------------------------------*/

import java.util.*;
import java.io.*;
import java.net.*;

/**
 *
 *
 * @author   Sverre H. Huseby
 *           &lt;<A HREF="mailto:shh@thathost.com">shh@thathost.com</A>&gt;
 * @version  $Id: WebPage.java,v 1.2 2000/09/20 19:39:24 sverrehu Exp $
 */
public final class WebPage
extends Observable
implements Runnable {
/*-----------------------------------------------------------------------+
 |  PRIVATE PART                                                         |
 +----------------------------------------------------------------------*/
    private static final int MAX_THREADS = 7;
    private static Object waitObject = new Object();
    private static int    numThreads = 0;

    private PropertyProvider prop;

    private URL url = null;
    /** The title, or something. */
    private String title;
    /** Personal annotation. */
    private String description;
    /** Page when last seen by user. */
    private SmarterChecksum csInitial = new SmarterChecksum();
    /** From previous compare. */
    private SmarterChecksum csLast = new SmarterChecksum();

    private byte status;

    private static Thread getThread(Runnable runnable) {
        synchronized (waitObject) {
            while (numThreads == MAX_THREADS) {
                try {
                    waitObject.wait();
                } catch (InterruptedException e) {
                }
            }
            ++numThreads;
        }
        Thread ret = new Thread(runnable, "refresh thread " + numThreads);
        ret.setDaemon(true);
        ret.setPriority(Thread.MIN_PRIORITY);
        return ret;

    }

    private static void returnThread(Thread t) {
        synchronized (waitObject) {
            if (--numThreads < 0)  /* test just in case... */
                numThreads = 0;
            waitObject.notify();
        }
        t = null;
    }

    private String extractTitle(String content) {
        String lc;
        int    idxStart, idxEnd;
        String tagStart = "<title>";
        String tagEnd   = "</title>";

        /* this won't always work, particularly if the title tags doesn't
         * look as expected. */
        lc = content.toLowerCase();
        if ((idxStart = lc.indexOf(tagStart)) < 0)
            return "";
        idxStart += tagStart.length();
        if ((idxEnd = lc.indexOf(tagEnd, idxStart)) < 0)
            return "";
        return content.substring(idxStart, idxEnd).trim();
    }

    private int getPage(StringBuffer ret) throws IOException {
        Properties sysProp;
        int ResponseCode = HttpURLConnection.HTTP_OK;// assume by default that it is ok even if not http

        BufferedReader br  = null;
        try {
            URLConnection con = url.openConnection();
            con.setDefaultUseCaches(false);
            con.setDoInput(true);
            con.setDoOutput(false);

            if (con instanceof HttpURLConnection) {
                HttpURLConnection httpcon = (HttpURLConnection) con;
                ResponseCode = httpcon.getResponseCode();
                if(ResponseCode != httpcon.HTTP_OK) {
//                      System.out.println("Bad response code "+ResponseCode);
                    httpcon.disconnect();
                    return(ResponseCode);
                }
                br = new BufferedReader(new InputStreamReader(
                                                    con.getInputStream()));
                String line;
                int NumberOfLines = 0;
                while ((line = br.readLine()) != null) {
                    ret.append(line + "\n");
                    ++NumberOfLines;
                }
//                  if(NumberOfLines == 0)
//                      System.out.println("Warning... Empty response!!!");
//                  if(NumberOfLines == 1)
//                      System.out.println("Warning... Single line!");
//                  System.out.println("Read "+NumberOfLines+" lines");
                httpcon.disconnect();
            } else {
                br = new BufferedReader(new InputStreamReader(
                                                    con.getInputStream()));
                String line;
                while ((line = br.readLine()) != null) {
                    ret.append(line + "\n");
                }
            }
        } catch (IOException e) {
        } finally {
            if(br != null)
                br.close();
        }
        return ResponseCode;
    }

    private int updateChecksum(StringBuffer sb)
    throws IOException {
        int ResponseCode = getPage(sb);

        csLast.calculate(sb);

        /* also set title, if not set before. */
        if (title == null || title.length() == 0)
            setTitle(extractTitle(sb.toString()));
        return(ResponseCode);
    }

    private void doRefreshThreadStuff() {
        int newStatus;
        /* if the previous check we did shows that the page has
         * changed, there is no need to check it again. */
        if (csInitial.isInitialized() && !csLast.equals(csInitial))
            setStatus(PAGE_UPDATED);
        else {
            /* now we know that we need to perform a new test. */
            configureProxy();
            try {
                StringBuffer sb = new StringBuffer();
                // we don't assume it is HTTP, but we use the same
                // codes for simplicity
                if (updateChecksum(sb) != HttpURLConnection.HTTP_OK) {
                    setStatus(ERROR);
                } else if (csLast.equals(csInitial))
                    setStatus(PAGE_NOT_UPDATED);
                else {
                    setStatus(PAGE_UPDATED);
                    DiskLog.log(this.title, sb.toString());
                }
            } catch (IOException e) {
                e.printStackTrace();
                setStatus(ERROR);
            }
        }
    }

    private void configureProxy() {
        Properties sysProp = System.getProperties();
        if (prop == null) {
            sysProp.put("http.proxyHost", prop.getProperty("proxy.host"));
            sysProp.put("http.proxyPort", prop.getProperty("proxy.port"));
        } else if (prop.getBooleanProperty("useProxy")) {
            sysProp.put("http.proxyHost", prop.getProperty("proxy.host"));
            sysProp.put("http.proxyPort", prop.getProperty("proxy.port"));
        } else {
            sysProp.remove("http.proxyHost");
            sysProp.remove("http.proxyPort");
        }
    }

    /* reading a line normally returns null on EOF */
    private String readLine(BufferedReader in)
    throws IOException {
        String ret;

        ret = in.readLine();
        if (ret == null)
            throw new EOFException("unexpected end of file reading URL file");
        return ret;
    }

    private void setStatus(byte status) {
        this.status = status;
        setChanged();
        notifyObservers();
    }

    private void setStatusWithoutNotify(byte status) {
        this.status = status;
    }

/*-----------------------------------------------------------------------+
 |  PACKAGE LOCAL PART                                                   |
 +----------------------------------------------------------------------*/
    static final int compare(WebPage x1, WebPage x2,
                             boolean changedOnTop, boolean ignoreCaseInSort) {
        if (changedOnTop) {
            if (x1.status == PAGE_UPDATED && x2.status != PAGE_UPDATED)
                return -1;
            if (x1.status != PAGE_UPDATED && x2.status == PAGE_UPDATED)
                return 1;
        }
        if (ignoreCaseInSort)
            return x1.getSomeTitle().compareToIgnoreCase(x2.getSomeTitle());
        return x1.getSomeTitle().compareTo(x2.getSomeTitle());
    }

    static final void sortArray(WebPage[] x, boolean changedOnTop,
                                boolean ignoreCaseInSort) {
        int     q, w, smallest;
        WebPage tmp;

        for (q = 0; q < x.length; q++) {
            smallest = q;
            for (w = q + 1; w < x.length; w++) {
                if (compare(x[w], x[smallest],
                            changedOnTop,ignoreCaseInSort) < 0)
                    smallest = w;
            }
            tmp = x[q];
            x[q] = x[smallest];
            x[smallest] = tmp;
        }
    }

    void load(BufferedReader in)
    throws IOException {
        String s;

        s = readLine(in);
        url = new URL(s);
        s = readLine(in);
        title = s;
        s = readLine(in);
        csInitial.setChecksum(s);
        s = readLine(in);
        csLast.setChecksum(s);
        description = "";
        for (;;) {
            s = readLine(in);
            if (s.equals("."))
                break;
            if (description.length() > 0)
                description += "\n";
            if (s.charAt(0) == '.')
                description += s.substring(1);
            else
                description += s;
        }
        if (csInitial.isInitialized() && !csLast.equals(csInitial))
            setStatusWithoutNotify(PAGE_UPDATED);
        else
            setStatusWithoutNotify(NO_RESULT);
    }

    void save(BufferedWriter out)
    throws IOException {
        String s;
        int    pos, next, len;

        if (url != null) {
            s = url.toString();
            out.write(s, 0, s.length());
        }
        out.newLine();
        if (title != null)
            out.write(title, 0, title.length());
        out.newLine();
        if (csInitial != null) {
            s = csInitial.toString();
            out.write(s, 0, s.length());
        }
        out.newLine();
        if (csLast != null) {
            s = csLast.toString();
            out.write(s, 0, s.length());
        }
        out.newLine();
        if (description != null) {
            for (pos = 0; pos < description.length(); pos = next + 1) {
                next = description.indexOf("\n", pos);
                if (next < 0)
                    next = description.length();
                len = next - pos;
                if (len > 0 || next < description.length()) {
                    if (description.charAt(pos) == '.')
                        out.write('.');
                    out.write(description, pos, len);
                    out.newLine();
                }
            }
        }
        out.write('.');
        out.newLine();
    }

/*-----------------------------------------------------------------------+
 |  PUBLIC INTERFACE                                                     |
 +----------------------------------------------------------------------*/
    public static final byte NO_RESULT        = 0;
    public static final byte QUEUED           = 1;
    public static final byte WORKING          = 2;
    public static final byte ERROR            = 3;
    public static final byte PAGE_NOT_UPDATED = 4;
    public static final byte PAGE_UPDATED     = 5;

    public WebPage(PropertyProvider prop) {
        this.prop = prop;

        setStatusWithoutNotify(NO_RESULT);
    }

    public WebPage(PropertyProvider prop, URL url) {
        this(prop);
        setURL(url);
    }

    public WebPage(PropertyProvider prop, String url)
    throws MalformedURLException {
        this(prop);
        setURL(url);
    }

    public URL getURL() {
        return url;
    }

    public void setURL(URL url) {
        if (!url.equals(this.url)) {
            this.url = url;
            setStatus(NO_RESULT);
        }
    }

    public void setURL(String url)
    throws MalformedURLException {
        setURL(new URL(url));
    }

    public String getTitle() {
        return title;
    }

    public void setTitle(String title) {
        if (this.title == null || title.compareTo(this.title) != 0) {
            this.title = title;
            setChanged();
            notifyObservers();
        }
    }

    public String getDescription() {
        return description;
    }

    public void setDescription(String description) {
        if (this.description == null
            || description.compareTo(this.description) != 0) {
            this.description = description;
            setChanged();
            notifyObservers();
        }
    }

    public String getSomeTitle() {
        if (title != null && title.length() > 0)
            return title;
        if (url != null)
            return url.toString();
        return "";
    }

    public void refresh() {
        Thread thread;

        synchronized (this) {
            if (status == WORKING || status == QUEUED)
                return;
            setStatus(QUEUED);
        }
        thread = getThread(this);
        setStatus(WORKING);
        thread.start();
    }

    public byte getStatus() {
        return status;
    }

    public String getStatusText() {
        switch (status) {
          case NO_RESULT:        return "Not yet checked";
          case QUEUED:           return "Queued for checking";
          case WORKING:          return "Checking...";
          case ERROR:            return "Couldn't access";
          case PAGE_NOT_UPDATED: return "Not changed since last visited";
          case PAGE_UPDATED:     return "Changed since last visited";
          default:               return "Internal error in WebPage.java";
        }
    }

    public void markAsSeen() {
        csInitial = csLast;
        setStatus(PAGE_NOT_UPDATED);
    }

    /* Runnable *********************************************************/
    public void run() {
        Thread.currentThread().setPriority(Thread.MIN_PRIORITY);

        try {
            doRefreshThreadStuff();
        } catch(Exception e) {
            e.printStackTrace();
        } finally {
            returnThread(Thread.currentThread());
        }
    }
}
